/************************CASP_COPYRIGHT**********************************
** Copyright (c) AadhunikLabs. All rights reserved.
** This file is part of CASP.
** Licensees holding valid licenses may use this file in accordance
** with the license agreement provided with the software.
*************************************************************************/
#include "fpga.h"
#include "sim_helper.h"
#ifndef UNDER_TESTING
#include "simcode.h"
#include "simcode_global.h"
#include "RootSignalClass.h"
#endif

static volatile bool g_system_initialized;
static volatile _UINT64 g_rt_clk_ticks;

#ifdef UNDER_TESTING
#ifdef FPGA_ENABLE_PCIE_COMM
#define COMM_CTRL_SIZE          PCIE_CTRL_SIZE
#define COMM_DATA_SIZE          PCIE_SHM_SIZE-COMM_CTRL_SIZE
#define COMM_BUF_SIZE           PCIE_SHM_SIZE
#elif defined FPGA_ENABLE_ETH_COMM
#define COMM_CTRL_SIZE          ETH_CTRL_SIZE
#define COMM_DATA_SIZE          ETH_SHM_SIZE-COMM_CTRL_SIZE
#define COMM_BUF_SIZE           ETH_SHM_SIZE
#elif defined FPGA_ENABLE_UART_COMM
#define COMM_CTRL_SIZE          UART_CTRL_SIZE
#define COMM_DATA_SIZE          UART_SHM_SIZE-COMM_CTRL_SIZE
#define COMM_BUF_SIZE           UART_SHM_SIZE
#elif defined FPGA_ENABLE_SHM_0
#define COMM_CTRL_SIZE          SHM0_CTRL_SIZE
#define COMM_DATA_SIZE          SHM0_DATA_SIZE
#define COMM_BUF_SIZE           COMM_DATA_SIZE+COMM_CTRL_SIZE
#endif
#define COMM_CTRL_RSIZE_IDX (COMM_BUF_SIZE-1)//for master
#define COMM_CTRL_WSIZE_IDX (COMM_BUF_SIZE-2)//for master
//
static AOUT_DTYPE a_o = 0;
static _BIT d_o = 0;
//
void setup(volatile _UINT32 *shm);
void loop(volatile _UINT32 *shm, AOUT_DTYPE &ao0, _BIT &do0);
void comm_init(volatile _UINT32 *shm);
_UINT32 comm_available(volatile _UINT32 *shm);
_INT32 comm_peek(_UINT32 idx, volatile _UINT32 *shm);//idx is not validated
void comm_peek_array(volatile _INT32 *buffer, _UINT32 size_to_read, volatile _UINT32 *shm);//no validations are done on size_to_read. starts from 0 index
void comm_read_clear(volatile _UINT32 *shm);
bool comm_write_idle(volatile _UINT32 *shm);
void comm_write(_INT32 data, _UINT32 offset_idx, volatile _UINT32 *shm);
void comm_write_array(volatile _INT32 *data, _UINT32 size, _UINT32 offset_idx, volatile _UINT32 *shm);
void comm_write_flush(_UINT32 wsize, volatile _UINT32 *shm);
void comm_write_clear(volatile _UINT32 *shm);
//
static _INT32 g_local_read_buf[24];
static _UINT16 g_write_size = 0;
#else
static volatile GLOBAL_RUN_PARAM_STRUCT g_run_para;
static RootSignalClass g_rootsig_obj;
#endif

void fpga_main(_UINT64 rt_clk_ticks
#ifdef FPGA_ENABLE_PCIE_COMM
        , volatile _UINT64 pcie_shm[PCIE_SHM_SIZE]
#endif
#ifdef FPGA_ENABLE_ETH_COMM
        , volatile _UINT32 eth_shm[ETH_SHM_SIZE]
#endif
#ifdef FPGA_ENABLE_UART_COMM
        , volatile _UINT32 uart_shm[UART_SHM_SIZE]
#endif
#ifdef FPGA_ENABLE_SHM_0
        , volatile _UINT32 shm0[SHM0_BUF_SIZE]
#endif
#ifdef FPGA_ENABLE_SHM_1
        , volatile _UINT32 shm1[SHM1_BUF_SIZE]
#endif
#ifdef FPGA_ENABLE_SHM_2
        , volatile _UINT32 shm2[SHM2_BUF_SIZE]
#endif
#ifdef FPGA_ENABLE_SHM_3
        , volatile _UINT32 shm3[SHM3_BUF_SIZE]
#endif
        , AIN_DTYPE ai0, _BIT di0, AOUT_DTYPE &ao0, _BIT &do0
        )
{
//#pragma HLS PIPELINE II=1
//#pragma HLS DATAFLOW
#pragma HLS RESET variable=g_system_initialized
#ifdef FPGA_ENABLE_ETH_COMM
#pragma HLS INTERFACE bram port=eth_shm
#pragma HLS RESOURCE variable=eth_shm core=RAM_1P_BRAM
#endif
//
#ifdef FPGA_ENABLE_UART_COMM
#pragma HLS INTERFACE bram port=uart_shm
#pragma HLS RESOURCE variable=uart_shm core=RAM_1P_BRAM
#endif
//
#ifdef FPGA_ENABLE_SHM_0
#ifdef FPGA_SHM0_AXI_MASTER
#pragma HLS INTERFACE m_axi depth=4196 port=shm0 offset=off
#elif defined FPGA_SHM0_AXI_LITE_SLAVE
#pragma HLS INTERFACE s_axilite port=shm0
#else
#pragma HLS INTERFACE bram port=shm0
#pragma HLS RESOURCE variable=shm0 core=RAM_1P_BRAM
#endif
#endif
//
#ifdef FPGA_ENABLE_SHM_1
#ifdef FPGA_SHM1_AXI_MASTER
#pragma HLS INTERFACE m_axi depth=4196 port=shm1 offset=off
#elif defined FPGA_SHM1_AXI_LITE_SLAVE
#pragma HLS INTERFACE s_axilite port=shm1
#else
#pragma HLS INTERFACE bram port=shm1
#pragma HLS RESOURCE variable=shm1 core=RAM_1P_BRAM
#endif
#endif
//
#ifdef FPGA_ENABLE_SHM_2
#ifdef FPGA_SHM2_AXI_MASTER
#pragma HLS INTERFACE m_axi depth=4196 port=shm2 offset=off
#elif defined FPGA_SHM2_AXI_LITE_SLAVE
#pragma HLS INTERFACE s_axilite port=shm2
#else
#pragma HLS INTERFACE bram port=shm2
#pragma HLS RESOURCE variable=shm2 core=RAM_1P_BRAM
#endif
#endif
//
#ifdef FPGA_ENABLE_SHM_3
#ifdef FPGA_SHM3_AXI_MASTER
#pragma HLS INTERFACE m_axi depth=4196 port=shm3 offset=off
#elif defined FPGA_SHM3_AXI_LITE_SLAVE
#pragma HLS INTERFACE s_axilite port=shm3
#else
#pragma HLS INTERFACE bram port=shm3
#pragma HLS RESOURCE variable=shm3 core=RAM_1P_BRAM
#endif
#endif
	//
        _UINT64 hw_time = micros();
	g_rt_clk_ticks = rt_clk_ticks;
	if(g_system_initialized == 0)
	{
#ifdef UNDER_TESTING
#ifdef FPGA_ENABLE_ETH_COMM
			setup(eth_shm);
#elif defined FPGA_ENABLE_UART_COMM
			setup(uart_shm);
#elif defined FPGA_ENABLE_SHM_0
			setup(shm0);
#endif
#else
            g_run_para.rt_clk_tick = hw_time;
            g_run_para.end_time = 0;
            g_run_para.sim_stepsize = SIM_TIME_STEP;
            g_run_para.plot_stepsize = SIM_PLOT_STEP-1;//to enable > instead of ==
            g_run_para.max_nsamples = 0;
            //
            g_run_para.time_step_0 = 0;
            g_run_para.time_step_1 = 0;
            g_run_para.time_step_plot = SIM_PLOT_STEP+1;
            g_run_para.pause_simulation = 0;
            g_run_para.stop_simulation = 0;
            g_run_para.user_paused = false;
            //
#ifdef _SIM_REAL_TIME
            g_run_para.prev_hw_time = 0;//hw_time;//let first step excute immediatly
#endif
            BLOCK_CONSTRUCTOR_PARAM_STRUCT cpara;
            cpara.nport_data = 0;
            cpara.blk_id = 0;
            g_rootsig_obj.Initialize(&cpara);
            for(int i=0; i<PRERUN_METHOD_CALL_COUNT; ++i)//number of passes
                g_rootsig_obj.PreRun(&g_run_para);
#endif
	}
        else if(g_run_para.user_paused == false)
	{
#ifdef UNDER_TESTING
		//ao0 = ai0;
		//do0 = di0;
#ifdef FPGA_ENABLE_ETH_COMM
		loop(eth_shm, ao0, do0);
#elif defined FPGA_ENABLE_UART_COMM
		loop(uart_shm, ao0, do0);
#elif defined FPGA_ENABLE_SHM_0
		loop(shm0, ao0, do0);
#endif//FPGA_ENABLE_ETH_COMM
#else//UNDER_TESTING
            g_run_para.rt_clk_tick = hw_time;
#if (SIM_TARGETHW_RUN_TIME > 0)
            if(g_run_para.rt_clk_tick > ((SIM_TARGETHW_RUN_TIME*1000)+1000000))
            {
                g_rootsig_obj.PostRun(&g_run_para);
                g_run_para.user_paused = true;
            }
#endif//SIM_TARGETHW_RUN_TIME
#ifdef _SIM_REAL_TIME
            //update current hw time
            _UINT64 hw_time_diff = 0;
            if(hw_time < g_run_para.prev_hw_time)
                hw_time_diff = (0xffffffff-g_run_para.prev_hw_time)+hw_time;
            else
                hw_time_diff = hw_time - g_run_para.prev_hw_time;//in usecs
            if(((_UINT64)g_run_para.sim_stepsize*(_UINT64)SIM_TIME_RESOLUTION) <= (_UINT64)hw_time_diff*(_UINT64)1000ULL)
            {
                g_run_para.prev_hw_time = hw_time;
#else
            {
#endif//_SIM_REAL_TIME
                g_rootsig_obj.Run(&g_run_para);
                if((_UINT64)g_run_para.time_step_plot > (_UINT64)g_run_para.plot_stepsize)
                {
                    //g_rootsig_obj.RunPlotStep(&g_run_para);
                    g_run_para.time_step_plot = 0;
                }
                g_run_para.time_step_1 = g_run_para.time_step_0;
                g_run_para.time_step_0 = g_run_para.time_step_0 + g_run_para.sim_stepsize;
                g_run_para.time_step_plot = g_run_para.time_step_plot + g_run_para.plot_stepsize;
            }
#if (SIM_TARGETHW_RUN_DELAY > 0)
            delay(SIM_TARGETHW_RUN_DELAY);
#endif
#endif//UNDER_TESTING
	}
	g_system_initialized = 1;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef UNDER_TESTING
void setup(volatile _UINT32 *shm)
{
	comm_init(shm);
    for(int i=0; i<24; ++i)
        g_local_read_buf[i] = 0;
    g_write_size = 0;
    //
    a_o = 0;
    d_o = 0;
}

void loop(volatile _UINT32 *shm, AOUT_DTYPE &ao0, _BIT &do0)
{
    _UINT32 rsize = comm_available(shm);
    fpga_wait();
    if(rsize == 5)
    {
        comm_peek_array(g_local_read_buf, rsize, shm);
        comm_read_clear(shm);
        comm_write_clear(shm);
    	//
        a_o = g_local_read_buf[1];
        d_o = g_local_read_buf[2];
    	//
    	g_write_size = rsize;
    }
    fpga_wait();
    if(comm_write_idle(shm) && g_write_size > 0)
    {
        comm_write_array(g_local_read_buf, g_write_size, 0, shm);
        comm_write_flush(g_write_size, shm);
        g_write_size = 0;
    }
    fpga_wait();
    //
    ao0 = a_o;
    do0 = d_o;
}
#endif //UNDER_TESTING
//////////////////////////////////////////////////////////////////////////////////////////////////////////
_UINT64 millis()
{
    //_UINT64 ticks = *g_rt_clk_ticks;
    //fpga_wait();
    return (g_rt_clk_ticks >> 21);//20 - divide by 1048576
}

_UINT64 micros()
{
    //_UINT64 ticks = *g_rt_clk_ticks;
    //fpga_wait();
    return (g_rt_clk_ticks >> 12);//10 - divide by 1024
}

void delay(_UINT64 delay)
{
    _QWORD start_time = millis();
    while(1)
    {
        if((millis() - start_time) > delay)
            break;
    }
}

void delayMicroseconds(_UINT64 delay)
{
    _QWORD start_time = micros();
    while(1)
    {
        if((micros() - start_time) > delay)
            break;
    }
}

void fpga_wait()
{
#ifdef _BUILD_CMP_VIVADO_HLS
    //wait();
    //minimum 2. by trail and error. otherwise 2 bram ports are created for uart.
    //recommended is 3. uart will be smooth.
    ap_wait_n(DELAY_WAIT_CLKS);
#endif
}
///////////////////////////////////////////////////comm////////////////////////////////////////////////////////////////////////
#ifdef UNDER_TESTING
void comm_init(volatile _UINT32 *shm)
{
    for(_UINT32 i=0; i<COMM_BUF_SIZE; ++i) shm[i] = 0;
    fpga_wait();
}

_UINT32 comm_available(volatile _UINT32 *shm)
{
    _UINT32 val = shm[COMM_CTRL_RSIZE_IDX];
    fpga_wait();
    return val;
}

_INT32 comm_peek(_UINT32 idx, volatile _UINT32 *shm)
{
    _INT32 *val = (_INT32*)&shm[idx];
    fpga_wait();
    return *val;
}

void comm_peek_array(volatile _INT32 *buffer, _UINT32 size_to_read, volatile _UINT32 *shm)
{
    for(_UINT32 i=0; i<size_to_read; ++i)
    {
    	_INT32 *val = (_INT32*)&shm[i];
    	buffer[i] = *val;
    }
    fpga_wait();
}

void comm_read_clear(volatile _UINT32 *shm)
{
    shm[COMM_CTRL_RSIZE_IDX] = 0;
    fpga_wait();
}

bool comm_write_idle(volatile _UINT32 *shm)
{
    _UINT32 val = shm[COMM_CTRL_WSIZE_IDX];
    fpga_wait();
    return (val > 0) ? 0 : 1;
}

void comm_write(_INT32 data, _UINT32 offset_idx, volatile _UINT32 *shm)
{
	_INT32 *val = (_INT32*)&shm[offset_idx];
	*val = data;
    fpga_wait();
}

void comm_write_array(volatile _INT32 *data, _UINT32 size, _UINT32 offset_idx, volatile _UINT32 *shm)
{
    for(_UINT32 i=0; i<size; ++i)
    {
    	_INT32 *val = (_INT32*)&shm[offset_idx+i];
    	*val = data[i];
    }
    fpga_wait();
}

void comm_write_flush(_UINT32 wsize, volatile _UINT32 *shm)
{
    shm[COMM_CTRL_WSIZE_IDX] = wsize;
    fpga_wait();
}

void comm_write_clear(volatile _UINT32 *shm)
{
    shm[COMM_CTRL_WSIZE_IDX] = 0;
    fpga_wait();
}
#endif
