/************************CASP_COPYRIGHT**********************************
** Copyright (c) AadhunikLabs. All rights reserved.
** This file is part of CASP.
** Licensees holding valid licenses may use this file in accordance
** with the license agreement provided with the software.
*************************************************************************/
//uart rx: 8 bits data, 1 start bit, 1 stop bit, no parity
//when receive is complete o_rx_dv will be driven high for one clock cycle
//UART_CLKS_PER_BIT = (CLK Frequency/(Baud Rate of UART)

`include "sim_common_definitions.vh"

`define TARGETHW_COMM_HEADER_SIZE 4 //shall be same as in parent program
`define PARENT_COMM_BUF_SIZE 128
`define PARENT_COMM_BUF_SIZE_BITS 8//(parent comm buffer size/2^n)*2
`define COMM_BUF_SIZE 64//shall be less than 256
`define UART_CLKS_PER_BIT 868 //=clk/baud. for 100MHz/115200=868, 50MHz/115200=434, 25MHz/115200=217, 10MHz/115200=87

`define COMM_READ_SIZE_IDX		0//in, set by child
`define COMM_READ_BUSY_IDX		1//out, set by parent
`define COMM_WRITE_SIZE_IDX		2//out, set by parent
`define COMM_WRITE_FREE_IDX		3//in, set by child

module uart_rx 
  (
   input        i_Clock,
   input        uart_rx_pin,
   output       o_Rx_DV,
   output [7:0] o_Rx_Byte
   );
    
  parameter s_IDLE         = 3'b000;
  parameter s_RX_START_BIT = 3'b001;
  parameter s_RX_DATA_BITS = 3'b010;
  parameter s_RX_STOP_BIT  = 3'b011;
  parameter s_CLEANUP      = 3'b100;
   
  reg           r_Rx_Data_R = 1'b1;
  reg           r_Rx_Data   = 1'b1;
   
  reg [15:0]    r_Clock_Count = 0;
  reg [2:0]     r_Bit_Index   = 0; //8 bits total
  reg [7:0]     r_Rx_Byte     = 0;
  reg           r_Rx_DV       = 0;
  reg [2:0]     r_SM_Main     = 0;
   
  // Purpose: Double-register the incoming data.
  // This allows it to be used in the UART RX Clock Domain.
  // (It removes problems caused by metastability)
  always @(posedge i_Clock)
    begin
      r_Rx_Data_R <= uart_rx_pin;
      r_Rx_Data   <= r_Rx_Data_R;
    end
   
   
  // Purpose: Control RX state machine
  always @(posedge i_Clock)
    begin
       
      case (r_SM_Main)
        s_IDLE :
          begin
            r_Rx_DV       <= 1'b0;
            r_Clock_Count <= 0;
            r_Bit_Index   <= 0;
             
            if (r_Rx_Data == 1'b0)          // Start bit detected
              r_SM_Main <= s_RX_START_BIT;
            else
              r_SM_Main <= s_IDLE;
          end
         
        // Check middle of start bit to make sure it's still low
        s_RX_START_BIT :
          begin
            if (r_Clock_Count == (`UART_CLKS_PER_BIT-1)/2)
              begin
                if (r_Rx_Data == 1'b0)
                  begin
                    r_Clock_Count <= 0;  // reset counter, found the middle
                    r_SM_Main     <= s_RX_DATA_BITS;
                  end
                else
                  r_SM_Main <= s_IDLE;
              end
            else
              begin
                r_Clock_Count <= r_Clock_Count + 1;
                r_SM_Main     <= s_RX_START_BIT;
              end
          end // case: s_RX_START_BIT
         
         
        // Wait UART_CLKS_PER_BIT-1 clock cycles to sample serial data
        s_RX_DATA_BITS :
          begin
            if (r_Clock_Count < `UART_CLKS_PER_BIT-1)
              begin
                r_Clock_Count <= r_Clock_Count + 1;
                r_SM_Main     <= s_RX_DATA_BITS;
              end
            else
              begin
                r_Clock_Count          <= 0;
                r_Rx_Byte[r_Bit_Index] <= r_Rx_Data;
                 
                // Check if we have received all bits
                if (r_Bit_Index < 7)
                  begin
                    r_Bit_Index <= r_Bit_Index + 1;
                    r_SM_Main   <= s_RX_DATA_BITS;
                  end
                else
                  begin
                    r_Bit_Index <= 0;
                    r_SM_Main   <= s_RX_STOP_BIT;
                  end
              end
          end // case: s_RX_DATA_BITS
     
     
        // Receive Stop bit.  Stop bit = 1
        s_RX_STOP_BIT :
          begin
            // Wait UART_CLKS_PER_BIT-1 clock cycles for Stop bit to finish
            if (r_Clock_Count < `UART_CLKS_PER_BIT-1)
              begin
                r_Clock_Count <= r_Clock_Count + 1;
                r_SM_Main     <= s_RX_STOP_BIT;
              end
            else
              begin
                r_Rx_DV       <= 1'b1;
                r_Clock_Count <= 0;
                r_SM_Main     <= s_CLEANUP;
              end
          end // case: s_RX_STOP_BIT
     
         
        // Stay here 1 clock
        s_CLEANUP :
          begin
            r_SM_Main <= s_IDLE;
            r_Rx_DV   <= 1'b0;
          end
         
         
        default :
          r_SM_Main <= s_IDLE;
         
      endcase
    end   
   
  assign o_Rx_DV   = r_Rx_DV;
  assign o_Rx_Byte = r_Rx_Byte;
   
endmodule //uart_rx

//uart tx: 8 bits data, 1 start bit, 1 stop bit, no parity
//when transmit is complete o_Tx_done will be driven high for one clock cycle
module uart_tx 
  (
   input       i_Clock,
   input       i_Tx_DV,
   input [7:0] i_Tx_Byte, 
   output reg  uart_tx_pin,
   output      o_Tx_Busy
   );
  
  parameter s_IDLE         = 3'b000;
  parameter s_TX_START_BIT = 3'b001;
  parameter s_TX_DATA_BITS = 3'b010;
  parameter s_TX_STOP_BIT  = 3'b011;
  parameter s_CLEANUP      = 3'b100;
   
  reg [2:0]    r_SM_Main     = 0;
  reg [15:0]   r_Clock_Count = 0;
  reg [2:0]    r_Bit_Index   = 0;
  reg [7:0]    r_Tx_Data     = 0;
  reg          r_Tx_Busy     = 1;
     
  always @(posedge i_Clock)
    begin
       
      case (r_SM_Main)
        s_IDLE :
          begin
            uart_tx_pin   <= 1'b1;         // Drive Line High for Idle
            r_Clock_Count <= 0;
            r_Bit_Index   <= 0;
            r_Tx_Busy     <= 0;
             
            if (i_Tx_DV == 1'b1)
              begin
                r_Tx_Busy   <= 1'b1;
                r_Tx_Data   <= i_Tx_Byte;
                r_SM_Main   <= s_TX_START_BIT;
              end
            else
              r_SM_Main <= s_IDLE;
          end // case: s_IDLE
         
         
        // Send out Start Bit. Start bit = 0
        s_TX_START_BIT :
          begin
            uart_tx_pin <= 1'b0;
             
            // Wait UART_CLKS_PER_BIT-1 clock cycles for start bit to finish
            if (r_Clock_Count < `UART_CLKS_PER_BIT-1)
              begin
                r_Clock_Count <= r_Clock_Count + 1;
                r_SM_Main     <= s_TX_START_BIT;
              end
            else
              begin
                r_Clock_Count <= 0;
                r_SM_Main     <= s_TX_DATA_BITS;
              end
          end // case: s_TX_START_BIT
         
         
        // Wait UART_CLKS_PER_BIT-1 clock cycles for data bits to finish         
        s_TX_DATA_BITS :
          begin
            uart_tx_pin <= r_Tx_Data[r_Bit_Index];
             
            if (r_Clock_Count < `UART_CLKS_PER_BIT-1)
              begin
                r_Clock_Count <= r_Clock_Count + 1;
                r_SM_Main     <= s_TX_DATA_BITS;
              end
            else
              begin
                r_Clock_Count <= 0;
                 
                // Check if we have sent out all bits
                if (r_Bit_Index < 7)
                  begin
                    r_Bit_Index <= r_Bit_Index + 1;
                    r_SM_Main   <= s_TX_DATA_BITS;
                  end
                else
                  begin
                    r_Bit_Index <= 0;
                    r_SM_Main   <= s_TX_STOP_BIT;
                  end
              end
          end // case: s_TX_DATA_BITS
         
         
        // Send out Stop bit.  Stop bit = 1
        s_TX_STOP_BIT :
          begin
            uart_tx_pin <= 1'b1;
             
            // Wait UART_CLKS_PER_BIT-1 clock cycles for Stop bit to finish
            if (r_Clock_Count < `UART_CLKS_PER_BIT-1)
              begin
                r_Clock_Count <= r_Clock_Count + 1;
                r_SM_Main     <= s_TX_STOP_BIT;
              end
            else
              begin
                r_Clock_Count <= 0;
                r_SM_Main     <= s_CLEANUP;
              end
          end // case: s_Tx_STOP_BIT
         
         
        // Stay here 1 clock
        s_CLEANUP :
          begin
            r_SM_Main <= s_IDLE;
          end
         
         
        default :
         begin
          r_SM_Main <= s_IDLE;
         end
      endcase
    end
  assign o_Tx_Busy   = r_Tx_Busy;
   
endmodule //uart_tx

/////////////////////////////////////////////////////uart_comm/////////////////////////////////////////////////////////
module uart_comm
(
input clk,
input comm_buf_we,
input comm_buf_ce,
input [`PARENT_COMM_BUF_SIZE_BITS-1:0] comm_buf_addr,//parent size may change based on buffer size
input [7:0] comm_buf_d0,
//
input comm_ctrl_we,
input comm_ctrl_ce,
input [1:0] comm_ctrl_addr,//parent size may change based on buffer size
input [15:0] comm_ctrl_buf_d0,
//
input UART_RXD,
//
output reg [7:0] comm_buf_q0,
output reg [15:0] comm_ctrl_buf_q0,
output UART_TXD
);

reg [7:0] recv_buf[`COMM_BUF_SIZE-1:0];
reg [7:0] uart_recv_temp_buf[15:0];//temp buffer
reg [7:0] send_buf[`COMM_BUF_SIZE-1:0];
reg [15:0] comm_write_bytes_pending = 0;
reg [15:0] comm_read_bytes_available = 0;
reg [15:0] comm_write_busy = 0;
reg [15:0] i_init;
initial begin
    comm_buf_q0 <= 0;
    comm_ctrl_buf_q0 <= 0;
    for(i_init=0; i_init<`COMM_BUF_SIZE; i_init=i_init+1) begin
        send_buf[i_init] <= 0;
        recv_buf[i_init] <= 0;
    end
end
//parent data exchange
always @ (posedge clk)
begin
    if(comm_buf_we) begin
        if(comm_write_bytes_pending == 0)
            send_buf[comm_buf_addr-`PARENT_COMM_BUF_SIZE] <= comm_buf_d0;//get data to send from parent
    end else if(comm_buf_ce) begin
        comm_buf_q0 <= recv_buf[comm_buf_addr];//send recvd. data to parent
    end
end
//status report to parent
always @ (posedge comm_ctrl_ce && comm_ctrl_we == 0)
begin
    if(comm_ctrl_addr == `COMM_READ_SIZE_IDX) begin
        comm_ctrl_buf_q0 <= comm_read_bytes_available;
    end else if(comm_ctrl_addr == `COMM_WRITE_FREE_IDX)
        comm_ctrl_buf_q0 <= comm_write_busy;
end

wire rx_valid;
wire [7:0] rx_data;
reg [5:0] rx_idx = 0;
reg comm_read_in_process = 0;
//read uart data to temp buffer and shift temp buffer data to main buffer. check no operations are done at uart and parent side during this period
reg [5:0] i_rx = 0;
//send data to serial port
reg tx_valid = 0;
reg [7:0] tx_data = 0;
reg [16:0] tx_idx = 0;//0-is invalid
wire tx_busy;

always @ (posedge clk)
begin
    if(rx_valid) begin
        uart_recv_temp_buf[rx_idx] = rx_data;
        rx_idx = rx_idx + 1;
        if(rx_idx > 15) rx_idx = 0;//overflow. control should never come here
    end
    if(comm_ctrl_we) begin
        if(comm_ctrl_addr == `COMM_READ_BUSY_IDX) begin
            if(comm_ctrl_buf_d0)
                comm_read_in_process <= 1;
            else begin
                comm_read_in_process <= 0;
                comm_read_bytes_available <= 0;
            end
        end
    end
    if(comm_read_in_process == 0) begin
        for(i_rx=0; i_rx<rx_idx; i_rx=i_rx+1) begin
            recv_buf[comm_read_bytes_available] = uart_recv_temp_buf[i_rx];
            comm_read_bytes_available = comm_read_bytes_available + 1;
        end
        rx_idx = 0;
    end
end

always @ (posedge clk)
begin
    //process commands from parent
    if(comm_ctrl_we) begin
        if(comm_ctrl_addr == `COMM_WRITE_SIZE_IDX) begin
            if(comm_write_bytes_pending == 0)
                comm_write_bytes_pending = comm_ctrl_buf_d0;
            if(comm_write_bytes_pending)
                comm_write_busy = 0;
        end
    end
    //write to uart
    if(tx_busy) begin
        tx_valid <= 0;
    end else begin
        if(comm_write_bytes_pending) begin
            tx_idx <= tx_idx + 1;
            if(tx_idx < comm_write_bytes_pending+1)//tx_idx = 0 is invalid
                tx_valid <= 1;
            else begin
                tx_valid <= 0;
                tx_idx <= 0;
                comm_write_bytes_pending <= 0;
            end
        end else begin
            comm_write_busy <= `COMM_BUF_SIZE;//put in else condition so that it waits for 1 cycle
        end
    end
    if(tx_idx > 0)
        tx_data <= send_buf[tx_idx-1];
    //tx_valid <= tx_valid;
end
//
uart_rx uart_rx_0(clk, UART_RXD, rx_valid, rx_data);
uart_tx uart_tx_0(clk, tx_valid, tx_data, UART_TXD, tx_busy);

endmodule //uart_comm
